from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP)
from qgis.core import (
  QgsCoordinateReferenceSystem,
  QgsProcessingParameterExtent,
  QgsProcessingParameterNumber,
  QgsProcessingParameterEnum,
  QgsProcessingParameterBoolean,
  QgsProcessingParameterFolderDestination,
  QgsProject
  )
from qgis import processing

from .fetchabstract import fetchabstract
import os
import glob

class fetchtengunextent(fetchabstract):
  
  # UIs
  PARAMETERS = {  
    "FETCH_EXTENT": {
      "ui_func": QgsProcessingParameterExtent,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("fetchtengunextent","Extent for fetching data")
      }
    },
    "MAX_NUM_FILES": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "type": QgsProcessingParameterNumber.Integer,
        "description": QT_TRANSLATE_NOOP("fetchtengunextent","Maximum number of files to be downloaded"),
        "defaultValue": 4
      }
    },       
    "PC_DATABASE": {
      "ui_func": QgsProcessingParameterEnum,
      "ui_args": {
        "options": [],
        "description": QT_TRANSLATE_NOOP("fetchtengunextent","Point cloud database"),
        "defaultValue": 0
      }
    },
    "ADD_PC_TO_CURRENT_PROJ": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchtengunextent","Add point cloud(s) to the current project?"),
        "defaultValue": True
      }
    },       
    "OUTPUT_FOLDER": {
      "ui_func": QgsProcessingParameterFolderDestination,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("fetchtengunextent","Folder to output")
      }
    }
  }
    
  CRS = ""
  OUTPUT_FOLDER = ""
    
  # initialization of the algorithm
  def initAlgorithm(self, config):    
    self.initUsingCanvas()
    self.PARAMETERS["PC_DATABASE"]["ui_args"]["options"] = [item["description"] for item in self.DBS]
    self.initParameters()
  
  
  def setWebFetchArgs(self, parameters, context, feedback):
    self.WEBFETCH_ARGS = {
      "REQUEST_ARGS": [],
      "SET": False,
      "DOWNLOADED_FILE": []
    }
    
    # set url and file names    
    tengun_list = processing.run(
      "tengungun:listtengun",
      {
        "FETCH_EXTENT": self.FETCH_AREA,
        "TARGET_CRS": self.CRS,
        "PC_DATABASE": [self.parameterAsEnum(parameters, "PC_DATABASE", context)],
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      feedback = feedback,
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    tengun_list = context.getMapLayer(tengun_list)
    
    
    for ft in tengun_list.getFeatures():
      self.WEBFETCH_ARGS["REQUEST_ARGS"].append(
        {
          "method": "GET",
          "url": ft["url"],
          "crs": QgsCoordinateReferenceSystem(ft["crs"])
        }
      )
    
    
    if len(self.WEBFETCH_ARGS["REQUEST_ARGS"]) > self.parameterAsInt(parameters, "MAX_NUM_FILES", context):
      feedback.reportError(self.tr("Number of files exceeds: ") + str(self.parameterAsInt(parameters, 'MAX_NUM_FILES', context)))
      raise Exception(self.tr("Number of files exceeds: ") + str(self.parameterAsInt(parameters, 'MAX_NUM_FILES', context)))
    
    if self.WEBFETCH_ARGS["REQUEST_ARGS"] is not None:
      self.WEBFETCH_ARGS["SET"] = True
    
  
  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback):        
    # import ptvsd
    # ptvsd.debug_this_thread()
    # set url and file names
    self.OUTPUT_FOLDER = self.parameterAsString(parameters, "OUTPUT_FOLDER", context)
    if len(self.OUTPUT_FOLDER) != len(self.OUTPUT_FOLDER.encode('utf-8')):
      feedback.reportError(self.tr("The output folder path must not contain multi-byte characters."))
      raise Exception(self.tr("The output folder path must not contain multi-byte characters."))
    if not os.path.exists(self.OUTPUT_FOLDER):
      os.makedirs(self.OUTPUT_FOLDER)
      
    # set url and file names
    self.CRS = QgsCoordinateReferenceSystem(self.DBS[self.parameterAsEnum(parameters, "PC_DATABASE", context)]["crs"])
    self.FETCH_AREA = self.parameterAsExtent(parameters, "FETCH_EXTENT", context, self.CRS)
    
    # set the meta information for obtaining SRTM data
    self.setWebFetchArgs(parameters, context, feedback)
    
    # download files using the session info
    self.fetchMultipleTenguns(parameters, context, feedback, parallel=False, output_dir=self.OUTPUT_FOLDER)
    
    for tng in glob.glob(os.path.join(self.OUTPUT_FOLDER, "*")):
      context.addLayerToLoadOnCompletion(
        tng, 
        context.LayerDetails(os.path.splitext(os.path.basename(tng))[0], QgsProject.instance(), "")
      )
    
    return {"OUTPUT": self.OUTPUT_FOLDER}
  
  def displayName(self):
    return self.tr("Fetch Point Cloud using extent")

  def createInstance(self):
    return fetchtengunextent()
