from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant)
from qgis.core import (
  QgsProcessingParameterFeatureSource,
  QgsProcessing,
  QgsProcessingParameterCrs,
  QgsProcessingParameterDistance,
  QgsProcessingParameterFeatureSink,
  QgsFeatureRequest,
  QgsFields,
  QgsField,
  QgsFeature,
  QgsWkbTypes,
  QgsVectorLayer,
  QgsProcessingParameterField
  )
from qgis import processing

from .algabstract import algabstract
import math

class transectlines(algabstract):
  
  # UIs
  PARAMETERS = {            
    "INPUT": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("transectlines","Line layer"),
        "types": [QgsProcessing.TypeVectorLine]
      }
    },
    "INPUT_ID_FIELD":{
      "ui_func": QgsProcessingParameterField,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","ID field of the line layer"),
        "parentLayerParameterName": "INPUT",
        "defaultValue": "PK",
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","Target CRS (Cartesian coordinates)")
      }
    },
    "INTERVAL": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","Interval of the transect lines"),
        "defaultValue": 25.0,
        "parentParameterName": "TARGET_CRS"
      }
    },      
    "MAXIMUM_INTERVAL": {
      "ui_func": QgsProcessingParameterDistance,
      "advanced": True,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","Max interval to reconstruct the original path"),
        "defaultValue": 25.0,
        "parentParameterName": "TARGET_CRS"
      }
    },      
    "LENGTH": {
      "ui_func": QgsProcessingParameterDistance,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","Length of the transect lines"),
        "defaultValue": 30.0,
        "parentParameterName": "TARGET_CRS"
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("transectlines","Transect lines")
      }
    }
  }
  
  FIELDS = {
    "transect_id":{"type": QVariant.Int}, # id of the transect line
    "transect_lid":{"type": QVariant.Int}, # id of the line, from which the transect line is created
    "topo_from_lid":{"type": QVariant.Int}, # id of the line, which is connected to the origin of the transect_lid
    "topo_from_vtype":{"type": QVariant.String}, # whether origin or terminal point is connected to the transect_lid
    "topo_to_lid":{"type": QVariant.Int}, # id of the line, which is connected to the terminal of the transect_lid
    "topo_to_vtype":{"type": QVariant.String}, # whether origin or terminal point is connected to the transect_lid
    "transect_sid":{"type": QVariant.Int}, # id of the segment of the line, from which the transect line is created
    "pnt_orig_x":{"type": QVariant.Double},
    "pnt_orig_y":{"type": QVariant.Double},
    # "angle_deg": {"type": QVariant.Double},
    # "length":    {"type": QVariant.Double}
  }
  
  LEN_TOLERANCE = 0.1
  
  # initialization of the algorithm
  def initAlgorithm(self, config):    
    (rect, target_crs) = self.getExtentAndCrsUsingCanvas()
    self.PARAMETERS["TARGET_CRS"]["ui_args"]["defaultValue"] = target_crs.authid()
    self.initParameters()
  
  # execution of the algorithm
  def processAlgorithm(self, parameters, context, feedback):        

      
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    # check whether the target CRS is x-y coordinates
    if target_crs.isGeographic() or target_crs.authid() == "EPSG:3857":
      feedback.reportError(self.tr("The Target CRS is NOT a Cartesian Coordinate System"))
      raise Exception(self.tr("The Target CRS is NOT a Cartesian Coordinate System"))
    
    # obtain line layer that is reprojected to the target CRS
    line_layer_raw = processing.run(
      "native:reprojectlayer", 
      {
        "INPUT": self.parameterAsSource(parameters, "INPUT", context).materialize(QgsFeatureRequest(), feedback),
        "TARGET_CRS": target_crs,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]      
                  
    # obtain single-part line layer
    line_layer_single = processing.run(
      "native:multiparttosingleparts",
      {
        "INPUT": line_layer_raw,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    line_layer_single = context.getMapLayer(line_layer_single)
    
    # prepare the fields
    path_flds = QgsFields()
    for fld in self.FIELDS:
      path_flds.append(QgsField(fld, self.FIELDS[fld]["type"]))
      
    # set sink and add features with values
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context,
      path_flds, QgsWkbTypes.LineString, target_crs
    )
    
    # prepare the transect intervals
    # if the distance is too large, shorter intervals are used (then the redundant transect lines are removed in the next step)
    transect_itvl = self.parameterAsDouble(parameters, "INTERVAL", context)
    max_itvl = self.parameterAsDouble(parameters, "MAXIMUM_INTERVAL", context)
    lid_fld = self.parameterAsString(parameters, "INPUT_ID_FIELD", context)
    transect_div = 1
    if transect_itvl > max_itvl:
      transect_div = math.ceil(transect_itvl / max_itvl)
      transect_itvl = transect_itvl / transect_div
    
    # obtain origin and terminal points of the line layer
    # which are used to identify the connection of the lines
    orig_term_pnts = processing.run(
      "native:extractspecificvertices",
      {
        "INPUT": line_layer_single,
        "VERTICES": "0,-1",
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    # assign primary key
    orig_term_pnts_lotid = processing.run(
      "native:addautoincrementalfield",
      {
        "INPUT": orig_term_pnts,
        "FIELD_NAME": "lotid",
        "START": 0,
        "GROUP_FIELDS": [],
        "SORT_ASCENDING": True,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    lotid_lid_dict = {}
    for ft_pnt in context.getMapLayer(orig_term_pnts_lotid).getFeatures():
      lotid_lid_dict[ft_pnt["lotid"]] = {
        "lid": ft_pnt[lid_fld],
        "vertex_type" : "origin" if ft_pnt["vertex_pos"] == 0 else "terminal"
      }
    
    # calculate the distance matrix
    dist_mat = processing.run(
      "qgis:distancematrix",
      {
        "INPUT": orig_term_pnts_lotid,
        "INPUT_FIELD": "lotid",
        "TARGET": orig_term_pnts_lotid,
        "TARGET_FIELD": "lotid",
        "MATRIX_TYPE": 0,
        "NEAREST_POINTS": 1,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    #delete features with large distance
    tol = 2.0
    dist_mat_ext = processing.run(
      "native:extractbyexpression",
      {
        "INPUT": dist_mat,
        "EXPRESSION": f"\"Distance\" < {tol}",
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm=True
    )["OUTPUT"]
    
    # identify which line is connected
    line_connections = {}
    for ft_dist_mat in context.getMapLayer(dist_mat_ext).getFeatures():
      ld = {
        "iid": lotid_lid_dict.get(ft_dist_mat["InputID"], {}).get("lid", -1) ,
        "from_or_to": "from" if lotid_lid_dict.get(ft_dist_mat["InputID"]).get("vertex_type", "unknown") =="origin" else "to",
        "tid": lotid_lid_dict.get(ft_dist_mat["TargetID"], {}).get("lid", -1),
        "tpnt_type": lotid_lid_dict.get(ft_dist_mat["InputID"]).get("vertex_type", "unknown"),
      }
            
      if line_connections.get(ld["iid"]) is None:
        line_connections[ld["iid"]] = {
          ld["from_or_to"] + "_lid": ld["tid"],
          ld["from_or_to"] + "_type": ld["tpnt_type"]
        }
      else:
        if line_connections[ld["iid"]].get(ld["from_or_to"] + "_lid") is None:
          line_connections[ld["iid"]][ld["from_or_to"] + "_lid"] = ld["tid"]
          line_connections[ld["iid"]][ld["from_or_to"] + "_type"] = ld["tpnt_type"]
        else:
          line_connections[ld["iid"]][ld["from_or_to"] + "_lid"] = -1
          line_connections[ld["iid"]][ld["from_or_to"] + "_type"] = "multiple"
          
  
      
      
    # iterate over the single line features
    # obtain the points on the line, create transect lines, and put them into the sink
    transect_id = 1
    for ft_line in line_layer_single.getFeatures():
      # extract single feature as a vector layer
      line_layer = QgsVectorLayer(f"LineString?crs={target_crs.authid()}", "temp", "memory")
      line_layer.dataProvider().addFeatures([ft_line])
      transect_lid = ft_line[lid_fld]
      line_connections_single = line_connections.get(transect_lid, {})
      
      # first, create points on the line
      pnts_on_line = processing.run(
        "native:pointsalonglines",
        {
          "INPUT": line_layer,
          "DISTANCE": transect_itvl,
          "START_OFFSET": 0,
          "END_OFFSET": 0,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm=True
      )["OUTPUT"]
      
      # connect the points, as a path
      path_from_pnts = processing.run(
        "native:pointstopath",
        {
          "INPUT": pnts_on_line,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm=True
      )["OUTPUT"]
      
      # make transect lines, which are at the vertices and perpendicular to the path 
      transect_paths = processing.run(
        "native:transect",
        {
          "INPUT": path_from_pnts,
          "LENGTH": self.parameterAsDouble(parameters, "LENGTH", context)*2,
          "ANGLE": 90,
          "SIDE": 2,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm=True
      )["OUTPUT"]
      
      # add origin information to the transect lines
      # transect_with_origin = QgsVectorLayer(f"LineString?crs={target_crs.authid()}&field=transect_id:integer&field=pnt_orig_x:double&field=pnt_orig_y:double", "temp", "memory")
      
      transect_sid = 0
      for i, (ft_pnt, ft_tr_line) in enumerate(zip(context.getMapLayer(pnts_on_line).getFeatures(), context.getMapLayer(transect_paths).getFeatures())):
        # if the transect line is not at the interval, skip
        if i % transect_div != 0:
          continue
        
        # create a new feature, if the transect line is at the interval
        ft_line_with_orig = QgsFeature(path_flds)
        ft_line_with_orig.setGeometry(ft_tr_line.geometry())
        ft_line_with_orig["transect_id"] = transect_id
        ft_line_with_orig["transect_lid"] = transect_lid
        ft_line_with_orig["transect_sid"] = transect_sid
        ft_line_with_orig["topo_from_lid"] = line_connections_single.get("from_lid", -1)
        ft_line_with_orig["topo_from_vtype"] = line_connections_single.get("from_type", "")
        ft_line_with_orig["topo_to_lid"] = line_connections_single.get("to_lid", -1)
        ft_line_with_orig["topo_to_vtype"] = line_connections_single.get("to_type", "")
        ft_line_with_orig["pnt_orig_x"] = ft_pnt.geometry().get().x()
        ft_line_with_orig["pnt_orig_y"] = ft_pnt.geometry().get().y()

        sink.addFeature(ft_line_with_orig)
        transect_id += 1
        transect_sid += 1
      
      transect_lid += 1
      
      # # cut of the transect lines at the crosssection of the line, if necessary
      # transect_paths_cut = processing.run(
      #   "native:splitwithlines",
      #   {
      #     "INPUT": transect_with_origin,
      #     "LINES": line_layer_single,
      #     "OUTPUT": "TEMPORARY_OUTPUT"
      #   },
      #   context = context,
      #   is_child_algorithm=True
      # )["OUTPUT"]
      
      # transect_paths_cut = context.getMapLayer(transect_paths_cut)
      
      # for path in transect_paths_cut.getFeatures():
      #   if path.geometry().length() >= self.parameterAsDouble(parameters, "LENGTH", context) - self.LEN_TOLERANCE:
      #     new_path = QgsFeature(path_flds)
          
      #     new_path["transect_id"] = transect_id
      #     transect_id += 1
          
      #     path_polyline = path.geometry().asPolyline()
      #     if abs(path_polyline[0].x() - path["pnt_orig_x"]) > self.LEN_TOLERANCE or abs(path_polyline[0].y() - path["pnt_orig_y"]) > self.LEN_TOLERANCE:
      #       path_polyline.reverse()
      #     new_path.setGeometry(QgsGeometry.fromPolylineXY(path_polyline))
            
      #     x0 = path_polyline[0].x()
      #     y0 = path_polyline[0].y()
      #     x1 = path_polyline[1].x()
      #     y1 = path_polyline[1].y()
          
      #     new_path["pnt_orig_x"] = x0
      #     new_path["pnt_orig_y"] = y0
          
      #     angle_deg = 450 - math.atan2(y1-y0, x1-x0) * 180 / math.pi # note! it is north clockwise
      #     angle_deg = angle_deg - 360.0 if angle_deg > 360.0 else angle_deg
          
      #     new_path["angle_deg"] = angle_deg
      #     new_path["length"] = path.geometry().length()

      #     sink.addFeature(new_path)
        
    return {"OUTPUT": dest_id}
  
  # Post processing; append layers
  def postProcessAlgorithm(self, context, feedback):
    return {}

  def displayName(self):
    return self.tr("Transect lines")

  def group(self):
    return self.tr('Tools')

  def groupId(self):
    return 'tools'

  def createInstance(self):
    return transectlines()
